#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>

#include <libadwin.h>

#include "cmdline.h"
#include "libadwin_test.h"
#include "libadwin/bits.h"

static struct gengetopt_args_info args_info;

typedef enum {
    INFO_DATA, INFO_FIFO,
} information_type_t;

typedef enum {
    VERBOSITY_LEVEL_NONE,
    VERBOSITY_LEVEL_QUIET,
    VERBOSITY_LEVEL_NORMAL,
    VERBOSITY_LEVEL_INFO,
    VERBOSITY_LEVEL_VERBOSE
} verbosity_level;

#define VL_QUIET   (args_info.verbose_arg >= VERBOSITY_LEVEL_QUIET)
#define VL_NORMAL  (args_info.verbose_arg >= VERBOSITY_LEVEL_NORMAL)
#define VL_INFO    (args_info.verbose_arg >= VERBOSITY_LEVEL_INFO)
#define VL_VERBOSE (args_info.verbose_arg >= VERBOSITY_LEVEL_VERBOSE)

#define MAX_DATA_SIZE 3000
#define EPSILON 0.00001

adwin_system_data_type_t adwin_system_data_types[] = {
    //  ADWIN_SYSTEM_SHORT,  // = 2, Transputer: 16 Bit, T9 and newer: 32 Bit
    ADWIN_SYSTEM_INTEGER,// = 3, Wordsize of processor, T2: 16 Bit, others: 32 Bit
    ADWIN_SYSTEM_LONG,   // = 4, 32 Bit
    ADWIN_SYSTEM_FLOAT,  // = 5, 32 Bit; not supported on T2 and T4
};
adwin_data_type_t adwin_data_types[] = {
    ADWIN_SHORT,  // = 1, 16 Bit
    ADWIN_LONG,   // = 2, 32 Bit
    ADWIN_FLOAT,  // = 5, 32 Bit
    ADWIN_DOUBLE, // = 6, 64 Bit
    ADWIN_VARIANT,// = 20, 128 Bit
};
long data_sizes[] = { 0, 1,
                      335, 336, 337, // packet size getDataSH = 336
                      349, 350, 351, // packet size Get_Data = 350
                      1749, 1750, 1751, // block size Get_Data = 1750
                      MAX_DATA_SIZE
                    };


typedef struct {
    int name;
    int size;
    information_type_t info_type;
    adwin_system_data_type_t data_type;
} adwin_data_info_t;

adwin_data_info_t datas[] = { {3, 3000, INFO_DATA, ADWIN_SYSTEM_SHORT},
    {4, 3000, INFO_DATA, ADWIN_SYSTEM_INTEGER},
    {5, 3000, INFO_DATA, ADWIN_SYSTEM_FLOAT},
    {13, 3000, INFO_FIFO, ADWIN_SYSTEM_SHORT},
    {14, 3000, INFO_FIFO, ADWIN_SYSTEM_INTEGER},
    {15, 3000, INFO_FIFO, ADWIN_SYSTEM_FLOAT},
    {40, 3000, INFO_DATA, ADWIN_SYSTEM_INTEGER}
};

int test_counter = 0;

/**
 * The last thing to call if an error occurred.
 */
void general_error()
{
    if (VL_NORMAL) {
        printf("%d tests successful, test %d failed\n",
               test_counter, test_counter + 1);
    }
    exit(-1);
}

/**
 * Somehow we received data different from what we expected.
 */
void data_error(double expected, double actual)
{
    if (VL_QUIET) {
        printf("failed.\n");
        printf("Data error, expected %f but received %f.\n", expected, actual);
    }
    general_error();
}

/**
 * Check if the ADwin function call resulted in an error.
 */
void check_error()
{
    int error_number = Get_Last_Error();

    if (error_number) {
        if (VL_QUIET) {
            printf("failed.\n");
            printf("Error message: %s\n", Get_Last_Error_Text(error_number));
        }
        general_error();
    } else {
        test_counter++;
    }
}

/**
 * The test was ok.
 */
void test_ok()
{
    if (VL_INFO)
        printf("ok.\n");
}

/**************************************************************
 * actual tests start here                                    *
 **************************************************************/

void test_Set_DeviceNo()
{
    if (VL_INFO)
        printf("testing Set_DeviceNo...");

    Set_DeviceNo(args_info.device_no_arg);
    check_error();
    test_ok();
}

const char* processor_to_btl(processor_type_t processor)
{
    switch (processor) {
        case PROCESSOR_T2:
            return "ADwin2.btl";
        case PROCESSOR_T4:
            return "ADwin4.btl";
        case PROCESSOR_T5:
            return "ADwin5.btl";
        case PROCESSOR_T8:
            return "ADwin8.btl";
        case PROCESSOR_T9:
            return "ADwin9.btl";
        case PROCESSOR_T10:
            return "ADwin10.btl";
        case PROCESSOR_T11:
            return "ADwin11.btl";
        case PROCESSOR_T12:
            return "ADwin12.btl";
        default:
            return "Unknown processor\n";
    }
}

processor_type_t int_to_processor(int processor_number)
{
    switch (processor_number) {
        case 2:
            return PROCESSOR_T2;
        case 4:
            return PROCESSOR_T4;
        case 5:
            return PROCESSOR_T5;
        case 8:
            return PROCESSOR_T8;
        case 9:
            return PROCESSOR_T9;
        case 10:
            return PROCESSOR_T10;
        case 11:
            return PROCESSOR_T11;
        case 12:
            return PROCESSOR_T12;
        default:
            return PROCESSOR_T9;
    }
}

void test_Boot()
{
    if (VL_INFO)
        printf("testing Boot...");

    processor_type_t processor =
        int_to_processor(args_info.processor_arg);

    Boot(processor_to_btl(processor));
    check_error();
    test_ok();
}

void test_Processor_Type()
{
    processor_type_t processor_reported, processor_specified;

    if (VL_INFO)
        printf("testing Processor_Type...");

    processor_specified = int_to_processor(args_info.processor_arg);
    processor_reported = Processor_Type();

    check_error();

    if (processor_specified != processor_reported)
        data_error(processor_specified, processor_reported);

    test_ok();
}

void test_Free_Mem()
{
    int32_t free_memory;

    if (VL_INFO)
        printf("testing Free_Mem...");

    if (args_info.processor_arg == 12) {
        free_memory = Free_Mem(MEM_UM);
        printf("free um : %d\n", free_memory);
        free_memory = Free_Mem(MEM_CM);
        printf("free cm : %d\n", free_memory);
        return;
    }

    if (args_info.processor_arg != 9) {
        if (VL_INFO)
            printf("NOT testing Free_Mem..."
                   "(proper values unknown)\n");

        return;
    }


    /* todo: T11, MEM_ALL, EM_LOCAL, MEM_ALL */
    free_memory = Free_Mem(MEM_PM_LOCAL);
    check_error();
    if (free_memory != 200958)
        data_error(200958, free_memory);

    free_memory = Free_Mem(MEM_DM_LOCAL);
    check_error();
    if (free_memory != 256160)
        data_error(256160, free_memory);

    free_memory = Free_Mem(MEM_DRAM_EXTERN);
    check_error();
    if (free_memory != 67108612)
        data_error(67108612, free_memory);

    test_ok();
}

void test_Load_Process()
{
    if (VL_INFO)
        printf("testing Load_Process...");

    processor_type_t processor = int_to_processor(args_info.processor_arg);

    switch (processor) {
        case PROCESSOR_T2:
            Load_Process("libadwin_test_process.t21");
            break;
        case PROCESSOR_T4:
            Load_Process("libadwin_test_process.t41");
            break;
        case PROCESSOR_T5:
            Load_Process("libadwin_test_process.t51");
            break;
        case PROCESSOR_T8:
            Load_Process("libadwin_test_process.t81");
            break;
        case PROCESSOR_T9:
            Load_Process("libadwin_test_process.t91");
            break;
        case PROCESSOR_T10:
            Load_Process("libadwin_test_process.ta1");
            break;
        case PROCESSOR_T11:
            Load_Process("libadwin_test_process.tb1");
            break;
        case PROCESSOR_T12:
            Load_Process("libadwin_test_process.tc1");
            break;
        default:
            general_error();
    }

    check_error();
    test_ok();
}


void test_Start_Process()
{
    if (VL_INFO)
        printf("testing Start_Process...");

    Start_Process(1);
    check_error();
    test_ok();
}

void test_Process_Status(int process_status)
{
    int reported_status;

    if (VL_INFO)
        printf("testing Process_Status %d...", process_status);

    reported_status = Process_Status(1);
    if ((process_status == 1) && (reported_status != 1))
        // process should be running, but is not
        data_error(process_status, reported_status);
    if ((process_status == 0) && (reported_status != 0))
        // process should be stopped, but is not
        data_error(process_status, reported_status);
    if ((process_status == 2) && (reported_status == 1))
        // process should be stopping, but is still running
        data_error(process_status, reported_status);

    check_error();
    test_ok();
}

void test_Stop_Process()
{
    if (VL_INFO)
        printf("testing Stop_Process...");

    Stop_Process(1);
    check_error();
    test_ok();
}

void test_Clear_Process()
{
    if (VL_INFO)
        printf("sleeping before Clear_Process...");

    sleep(2);
    test_ok();

    if (VL_INFO)
        printf("testing Clear_Process...");

    Clear_Process(1);
    check_error();
    test_ok();
}

void test_Set_Globaldelay(long delay)
{
    if (VL_INFO)
        printf("testing Set_Globaldelay %ld...", delay);

    Set_Globaldelay(1, delay);
    check_error();
    test_ok();
}

void test_Get_Globaldelay(long delay)
{
    long reported_delay;

    if (VL_INFO)
        printf("testing Get_Globaldelay %ld...", delay);

    reported_delay = Get_Globaldelay(1);
    if (reported_delay != delay)
        data_error(delay, reported_delay);

    check_error();
    test_ok();
}

void test_Set_Par()
{
    int i;

    if (VL_INFO)
        printf("testing Set_Par...");

    for (i = 1; i <= 80; i++) {
        Set_Par(i, i * i);
        check_error();
    }
    test_ok();
}

void test_Get_Par()
{
    int i;
    int32_t par;

    if (VL_INFO)
        printf("testing Get_Par...");

    for (i = 1; i <= 80; i++) {
        par = Get_Par(i);
        check_error();
        if (par != i * i) {
            data_error(i * i, par);
        }
    }
    test_ok();
}

void test_Set_FPar()
{
    int i;

    if (VL_INFO)
        printf("testing Set_FPar...");

    for (i = 1; i <= 80; i++) {
        Set_FPar(i, (float)i * i);
        check_error();
    }
    test_ok();
}

void test_Get_FPar()
{
    int i;
    float fpar;

    if (VL_INFO)
        printf("testing Get_FPar...");

    for (i = 1; i <= 80; i++) {
        fpar = Get_FPar(i);
        check_error();
        if (fpar - i * i > 0.00001) {
            data_error(i * i, fpar);
        }
    }
    test_ok();
}

void test_Data_Length()
{
    int i, tmp;

    if (VL_INFO)
        printf("testing Data_Length...");

    for (i = 0; i < sizeof(datas) / sizeof(datas[0]); i++) {
        tmp = Data_Length(datas[i].name);
        printf("%d %d %d\n", i, datas[i].size, tmp);

        check_error();
        if (tmp == datas[i].size)  // this is OK
            continue;

        if (tmp == datas[i].size + 2) // this is OK if data is string
            continue;


#if 0
        if (int_to_processor(args_info.processor_arg) == PROCESSOR_T12) {
            printf("%d %d\n", datas[i].size, tmp);
            if ((tmp >= datas[i].size) && (tmp <= datas[i].size + 3))
                continue;
        }

        if (int_to_processor(args_info.processor_arg) == PROCESSOR_T11) {
            // quad alignment for T11
            if ((tmp >= datas[i].size) && (tmp <= datas[i].size + 3))
                continue;
        }
#endif
        data_error(datas[i].size, tmp);
    }

    test_ok();
}

void test_SetData_Long(int size)
{
    int i;
    int32_t *src;

    if (VL_INFO)
        printf("testing SetData_Long %d...", size);

    src = calloc(size, sizeof(int32_t));
    for (i = 0; i < size; i++) {
        src[i] = i * i; // don't care about overflows
    }
    SetData_Long(ADWIN_SYSTEM_LONG, src, 1, size);
    for (i = 0; i < size; i++) {
        if (src[i] != i * i) {
            data_error(i * i, src[i]);
        }
    }
    free(src);
    check_error();
    test_ok();
}

void test_GetData_Long(int size)
{
    int i;
    int32_t *dest;

    if (VL_INFO)
        printf("testing GetData_Long %d...", size);

    dest = calloc(size, sizeof(int32_t));
    GetData_Long(ADWIN_SYSTEM_LONG, dest, 1, size);
    check_error();
    for (i = 0; i < size; i++) {
        if (dest[i] != i * i) {
            data_error(i * i, dest[i]);
        }
    }
    free(dest);
    test_ok();
}

void test_SetData_Float(int size)
{
    int i;
    float *src;

    if (VL_INFO)
        printf("testing SetData_Float %d...", size);

    src = calloc(size, sizeof(float));
    for (i = 0; i < size; i++) {
        src[i] = i * i; // don't care about overflows
    }
    SetData_Float(ADWIN_SYSTEM_FLOAT, src, 1, size);
    for (i = 0; i < size; i++) {
        if (src[i] != i * i) {
            data_error(i * i, src[i]);
        }
    }
    free(src);
    check_error();
    test_ok();
}

void test_GetData_Float(int size)
{
    int i;
    float *dest;

    if (VL_INFO)
        printf("testing GetData_Float %d...", size);

    dest = calloc(size, sizeof(float));
    GetData_Float(ADWIN_SYSTEM_FLOAT, dest, 1, size);
    check_error();
    for (i = 0; i < size; i++) {
        if (dest[i] != i * i) {
            data_error(i * i, dest[i]);
        }
    }
    free(dest);
    test_ok();
}

void test_Fifo_Empty(int fifo_no, int expect)
{
    int32_t empty;

    if (VL_INFO)
        printf("testing Fifo_Empty...");

    empty = Fifo_Empty(fifo_no);
    check_error();
    if ((empty != expect - 1) && (empty != expect)) {
        data_error(expect - 1, empty);
    }
    test_ok();
}

void test_Fifo_Full(int fifo_no, int expect)
{
    int32_t full;

    if (VL_INFO)
        printf("testing Fifo_Full...");

    full = Fifo_Full(fifo_no);
    check_error();
    if (full != expect) {
        data_error(expect, full);
    }
    test_ok();
}

void test_Fifo_Clear(int fifo_no)
{
    if (VL_INFO)
        printf("testing Fifo_Clear...");

    Fifo_Clear(fifo_no);
    check_error();
    test_ok();
}

void test_SetFifo_Long(int fifo_no, int size)
{
    int i;
    int32_t *src;

    if (VL_INFO)
        printf("testing SetFifo_Long %d...", size);

    src = calloc(size, sizeof(int32_t));
    printf("calloced %zu bytes at %p\n", size * sizeof(int32_t), src);
    for (i = 0; i < size; i++) {
        src[i] = i * i; // don't care about overflows
    }
    SetFifo_Long(fifo_no, src, size);
    free(src);
    check_error();
    test_ok();
}

void test_GetFifo_Long(int fifo_no, int size)
{
    int i;
    int32_t *dest;

    if (VL_INFO)
        printf("testing GetFifo_Long %d...", size);

    dest = calloc(size, sizeof(int32_t));
    GetFifo_Long(fifo_no, dest, size);
    check_error();
    for (i = 0; i < size; i++) {
        if (dest[i] != i * i) {
            data_error(i * i, dest[i]);
        }
    }
    free(dest);
    test_ok();
}

void debug_print_buffer(const void *buf, int32_t buf_len)
{
    int32_t i;

    for (i = 0; i < buf_len; i++) {
        printf("%02X ", ((unsigned char *)buf)[i]);
        if (i % 8 == 7)
            printf("\n");
    }
    printf("\n");
}



void test_Set_Info(information_type_t info, adwin_system_data_type_t asdt,
                   adwin_data_type_t adt, long count)
{
    int i;
    void *src;

    if (VL_INFO)
        printf("testing Set_%s adt=%d asdt=%d count=%ld...",
               (info == INFO_DATA ? "Data" : "Fifo"),
               adt, asdt, count);

    switch (adt) {
        case ADWIN_SHORT: {
            int16_t *tmp;
            tmp = calloc(count, sizeof(int16_t));
            for (i = 0; i < count; i++) {
                tmp[i] = i * i; // don't care about overflows
            }
            src = tmp;
            break;
        }
        case ADWIN_LONG: {
            int32_t *tmp;
            tmp = calloc(count, sizeof(int32_t));
            for (i = 0; i < count; i++) {
                tmp[i] = i * i; // don't care about overflows
            }
            src = tmp;
            break;
        }
        case ADWIN_FLOAT: {
            float *tmp;
            tmp = calloc(count, sizeof(float));
            for (i = 0; i < count; i++) {
                tmp[i] = i * i; // don't care about overflows
            }
            src = tmp;
            break;
        }
        case ADWIN_DOUBLE: {
            double *tmp;
            tmp = calloc(count, sizeof(double));
            for (i = 0; i < count; i++) {
                tmp[i] = i * i; // don't care about overflows
            }
            src = tmp;
            break;
        }
        case ADWIN_VARIANT: {
            struct myVariant *tmp;
            tmp = calloc(count, sizeof(struct myVariant));
            for (i = 0; i < count; i++) {
                // store values in all different formats...
                /* tmp[i].type = i % 4 + 2; */
                tmp[i].type = i % 3 + 3;
                switch (tmp[i].type) {
                    case VARIANT_SHORT: {
                        // not reached;
                        int16_t *p = (void *)tmp[i].value;
                        *p = (int16_t)(i * i);
                        break;
                    }
                    case VARIANT_LONG: {
                        int32_t *p = (void *)tmp[i].value;
                        *p = (int32_t)(i * i);
                        break;
                    }
                    case VARIANT_FLOAT: {
                        float *p = (void *)tmp[i].value;
                        *p = (float)(i * i);
                        break;
                    }
                    case VARIANT_DOUBLE: {
                        double *p = (void *)tmp[i].value;
                        *p = (double)(i * i);
                        break;
                    }
                    default:
                        if (VL_QUIET)
                            printf("Type problem Set_Data: %d.\n",
                                   tmp[i].type);
                        general_error();
                        break;
                }
            }
            src = tmp;
            break;
        }
        default:
            general_error();
            return;
    }
    switch (info) {
        case INFO_DATA:
            _Set_Data(asdt, src, adt, 1, count, args_info.device_no_arg, 1);
            break;
        case INFO_FIFO:
            _Set_Fifo(asdt + 10, src, adt, count, args_info.device_no_arg);
            break;
    }
    free(src);
    check_error();
    test_ok();
}

void test_Get_Info(information_type_t info, adwin_system_data_type_t asdt,
                   adwin_data_type_t adt, long count)
{
    int i;
    void *dest;

    if (VL_INFO)
        printf("testing Get_%s adt=%d asdt=%d count=%ld...",
               (info == INFO_DATA ? "Data" : "Fifo"),
               adt, asdt, count);

    switch (adt) {
        case ADWIN_SHORT:
            dest = calloc(count, sizeof(int16_t));
            break;
        case ADWIN_LONG:
            dest = calloc(count, sizeof(int32_t));
            break;
        case ADWIN_FLOAT:
            dest = calloc(count, sizeof(float));
            break;
        case ADWIN_DOUBLE:
            dest = calloc(count, sizeof(double));
            break;
        case ADWIN_VARIANT:
            dest = calloc(count, sizeof(struct myVariant));
            break;
        default:
            general_error();
            return;
    }
    switch (info) {
        case INFO_DATA:
            _Get_Data(asdt, dest, adt, 1, count, args_info.device_no_arg, 1);
            break;
        case INFO_FIFO:
            _Get_Fifo(asdt + 10, dest, adt, count, args_info.device_no_arg);
            break;
    }
    switch (adt) {
        case ADWIN_SHORT: {
            int16_t *tmp = dest;
            for (i = 0; i < count; i++) {
                if (tmp[i] != (int16_t)(i * i))
                    data_error((int16_t)(i * i), tmp[i]);
            }
            break;
        }
        case ADWIN_LONG: {
            int32_t *tmp = dest;
            for (i = 0; i < count; i++) {
                if (tmp[i] != (int32_t)(i * i))
                    data_error((int32_t)(i * i), tmp[i]);
            }
            break;
        }
        case ADWIN_FLOAT: {
            float *tmp = dest;
            for (i = 0; i < count; i++) {
                if (tmp[i] - i * i > EPSILON)
                    data_error((float)(i * i), tmp[i]);
            }
            break;
        }
        case ADWIN_DOUBLE: {
            double *tmp = dest;
            for (i = 0; i < count; i++) {
                if (tmp[i] - i * i > EPSILON)
                    data_error((double)(i * i), tmp[i]);
            }
            break;
        }
        case ADWIN_VARIANT: {
            struct myVariant *tmp = dest;
            for (i = 0; i < count; i++) {
                switch (tmp[i].type) {
                    case VARIANT_SHORT: {
                        int16_t *p = (void *)tmp[i].value;
                        if (*p != (int16_t)(i * i))
                            data_error((int16_t)(i * i), *p);
                        break;
                    }
                    case VARIANT_LONG: {
                        int32_t *p = (void *)tmp[i].value;
                        if (*p != (int32_t)(i * i))
                            data_error((int32_t)(i * i), *p);
                        break;
                    }
                    case VARIANT_FLOAT: {
                        float *p = (void *)tmp[i].value;
                        if (*p != (float)(i * i))
                            data_error((float)(i * i), *p);
                        break;
                    }
                    case VARIANT_DOUBLE: {
                        double *p = (void *)tmp[i].value;
                        if (*p != (double)(i * i))
                            data_error((double)(i * i), *p);
                        break;
                    }
                    default:
                        if (VL_QUIET)
                            printf("Unknown variant type %d.\n",
                                   tmp[i].type);
                        general_error();
                        break;
                }

            }
            break;
        }
        default:
            general_error();
            return;
    }
    free(dest);
    check_error();
    test_ok();
}

void test_Data(adwin_system_data_type_t asdt, adwin_data_type_t adt, long count)
{
    // data_X has adwin_system_data_type X
    test_Set_Info(INFO_DATA, asdt, adt, count);
    test_Get_Info(INFO_DATA, asdt, adt, count);
}

void test_Fifo(adwin_system_data_type_t asdt, adwin_data_type_t adt, long count)
{
    // data_10+X has adwin_system_data_type X

    int i, correct_fifo_size = -1;
    for (i = 0; i < sizeof(datas) / sizeof(datas[0]); i++) {
        if (datas[i].name == asdt + 10) {
            correct_fifo_size = datas[i].size;
        }
    }

    test_Fifo_Clear(10 + asdt);
    test_Fifo_Empty(10 + asdt, correct_fifo_size);
    test_Fifo_Full(10 + asdt, 0);

    test_Set_Info(INFO_FIFO, asdt, adt, count);

    test_Fifo_Empty(10 + asdt, correct_fifo_size - count);
    test_Fifo_Full(10 + asdt, count);

    test_Get_Info(INFO_FIFO, asdt, adt, count);

    test_Fifo_Empty(10 + asdt, correct_fifo_size);
    test_Fifo_Full(10 + asdt, 0);
}

void test_SetData_String(long count)
{
    long i;
    char *text;

    if (VL_INFO)
        printf("testing SetData_String %ld...", count);

    text = malloc(count + 1);
    for (i = 0; i < count; i++) {
        text[i] = i % 255 + 1;
    }
    text[count] = '\0';

    SetData_String(40, text);
    free(text);
    check_error();
    test_ok();
}

void test_GetData_String(long count)
{
    long i;
    char *text;

    if (VL_INFO)
        printf("testing GetData_String %ld...", count);

    text = malloc(count + 1);
    GetData_String(40, text, count + 1);
    check_error();

    for (i = 0; i < count; i++) {
        if (text[i] != (char)(i % 255 + 1)) {
            data_error((char)(i % 255 + 1), text[i]);
        }
    }
    free(text);
    test_ok();
}

void test_String_Length(long count)
{
    long len;

    if (VL_INFO)
        printf("testing String_Length %ld...", count);

    len = String_Length(40);
    check_error();
    if (len != count)
        data_error(count, len);

    test_ok();
}


void run_all_tests()
{
    int i, j, k;

    test_Set_DeviceNo();
    test_Boot();
    test_Processor_Type();
    // todo: test_Workload();
    test_Free_Mem();
    test_Load_Process();
    test_Start_Process();

    test_Set_Par();
    test_Get_Par();
    test_Set_FPar();
    test_Get_Par();

    test_Process_Status(1);
    test_Stop_Process();
    test_Process_Status(2);
    test_Clear_Process();
    test_Process_Status(0);

    test_Get_Globaldelay(40000000);
    test_Set_Globaldelay(0x7fffffff);
    test_Get_Globaldelay(0x7fffffff);

    test_Data_Length();

    // iterate over lengths
    for (i = 0; i < ARRAY_LENGTH(data_sizes); i++) {
        // iterate over data types on ADwin system
        for (j = 0; j < ARRAY_LENGTH(adwin_system_data_types); j++) {
            // iterate over local data types
            for (k = 0; k < ARRAY_LENGTH(adwin_data_types); k++) {
                test_Boot();
                test_Load_Process();
                test_Start_Process();
                test_Data(adwin_system_data_types[j], adwin_data_types[k], data_sizes[i]);
                test_Fifo(adwin_system_data_types[j], adwin_data_types[k], data_sizes[i]);
            }
        }
    }

    // miscellaneous tests:

    for (i = 0; i < ARRAY_LENGTH(data_sizes); i++) {
        test_Boot();
        test_Load_Process();
        test_Start_Process();

        test_SetData_Long(data_sizes[i]);
        test_GetData_Long(data_sizes[i]);

        test_SetData_Float(data_sizes[i]);
        test_GetData_Float(data_sizes[i]);

        test_SetData_String(data_sizes[i]);
        test_GetData_String(data_sizes[i]);
        test_String_Length(data_sizes[i]);
    }
}

int main(int argc, char **argv)
{
    int i;

    if (cmdline_parser(argc, argv, &args_info) != 0)
        return EXIT_FAILURE;

    // for T11, update fifo sizes.
    if (int_to_processor(args_info.processor_arg) == PROCESSOR_T11) {
        for (i = 0; i < sizeof(datas) / sizeof(datas[0]); i++)
            if (datas[i].info_type == INFO_FIFO)
                datas[i].size = datas[i].size / 4 * 4 + 3;
    }

    // for T12, update fifo sizes.
    if (int_to_processor(args_info.processor_arg) == PROCESSOR_T12) {
        for (i = 0; i < sizeof(datas) / sizeof(datas[0]); i++) {
            //if (datas[i].info_type == INFO_DATA) {
            datas[i].size++;
            //}
            //printf("expected data[%d]:  %d\n", datas[i].name, datas[i].size);
        }
    }

    for (i = 0; i < args_info.loops_arg; i++)
        run_all_tests();

    if (VL_NORMAL)
        printf("%d tests successful\n", test_counter);

    return 0;
}
